
class RotationType:
    RT_WHD = 0
    RT_HWD = 1
    RT_HDW = 2
    RT_DHW = 3
    RT_DWH = 4
    RT_WDH = 5

    ALL = [RT_WHD, RT_HWD]

class NoRotationType:
    RT_WHD = 0
    RT_HWD = 1
    RT_HDW = 2
    RT_DHW = 3
    RT_DWH = 4
    RT_WDH = 5

    ALL = [RT_WHD]


class Axis:
    WIDTH = 0
    HEIGHT = 1
    DEPTH = 2

    ALL = [WIDTH, HEIGHT, DEPTH]




from typing import List
from ..options import (
	DIMENSION_UNIT_OF_MEASURE_OPTIONS,
	WEIGHT_UNIT_OF_MEASURE_OPTIONS,
)
from ..unit_converters import (
	DIMENSION_CONVERTER,VOLUME_CONVERTER,WEIGHT_CONVERTER,
)
from ..py3dbp import Item
from .. import options
import numpy as np
from matplotlib import pyplot


def get_rectangle_plot_lists(ll : List, ur : List, x_padding : float, y_padding : float):
	'''
	Helper Function for Plotting Logistics Objects

	Parameters
	-----------
	ll : List
		List or array of length 2 describing the lower-left corner of the box to be plotted
	ur : List
		List or array of length 2 describing the upper-right corner of the box to be plotted
	x_padding : float
		Padding around the box
	y_padding : float
		Padding around the box
	
	Returns
	----------
	x_list : List
		List of 5 x values to create plot path for the box
	y_list : List
		List of 5 y values to create plot path for the box
	'''
	x_list = [
		ll[0] - x_padding,
		ur[0] + x_padding,
		ur[0] + x_padding,
		ll[0] - x_padding,
		ll[0] - x_padding
	]
	y_list = [
		ll[1] - y_padding,
		ll[1] - y_padding,
		ur[1] + y_padding,
		ur[1] + y_padding,
		ll[1] - y_padding
	]
	return x_list,y_list


class LogisticsObject:
	'''
	Generalized Logistics Object:  Piece, Shipment, or Trailer

	A generic class that handles length, width, and height dimensions
	as well as weight, along with their units.  This is not intended 
	to be a stand-alone object, it is intended to be inherited by 
	other objects to provide basic measurement and weight functionality.
	'''

	default_dimension_units = options.DimUomInches
	default_volume_units = options.VolumeUomCubicFeet
	default_weight_units = options.WeightUomPounds

	def __init__(self,length,height,width,dimension_unit_of_measure,weight,weight_unit_of_measure,**kwargs):
        # Ensure unit of measure are recognized
		assert dimension_unit_of_measure in list(map(lambda x : x[0],DIMENSION_UNIT_OF_MEASURE_OPTIONS))
		assert weight_unit_of_measure in list(map(lambda x : x[0],WEIGHT_UNIT_OF_MEASURE_OPTIONS))
        
        # Set raw and normalized dimensions
		self.dimension_unit_of_measure = dimension_unit_of_measure
		self.length_raw = length
		self.width_raw = width
		self.height_raw = height
		
		self.length = DIMENSION_CONVERTER[dimension_unit_of_measure][self.default_dimension_units](length)
		self.width = DIMENSION_CONVERTER[dimension_unit_of_measure][self.default_dimension_units](width)
		self.height = DIMENSION_CONVERTER[dimension_unit_of_measure][self.default_dimension_units](height)

        # Set raw and normalized weight
		self.weight_unit_of_measure = weight_unit_of_measure
		self.weight_raw = weight
		self.weight = WEIGHT_CONVERTER[weight_unit_of_measure][self.default_weight_units](weight)
        
        # Initialized position as 0,0,0
        # A frame of reference is required to interpret position attribute further
		# The position vector is of the form (length,width,height)
		self.position = np.array([0,0,0])
		self.set_var_types()
		self.is_rotated = False

		# Optional attributes, not used for optimization but for individual object references
		self.id = kwargs.get('id')
		self.name = kwargs.get('name')
		self.desc = kwargs.get('desc')
	
	def set_var_types(self,variables=None,dtype=float):
		'''
		Set Provided Variable Types
		'''
		if isinstance(variables,str):
			variables = [variables]
		if variables is None:
			variables = ['length','width','height','weight','length_raw','width_raw','height_raw','weight_raw']
		for var in variables:
			setattr(self,var,dtype(getattr(self,var)))
			
	def get_volume(self,units=None):
		'''
		Get the Volume of the Logistics Object
		'''
		volume = self.length * self.width * self.height
		if units is None:
			return VOLUME_CONVERTER[f'CUBIC_{self.default_dimension_units}'][self.default_volume_units](volume)
		return VOLUME_CONVERTER[self.default_volume_units][units](volume)
	
	def get_boundaries(self):
		'''
		Get the Profile of the Shipping Object
		'''
		boundaries = np.array([
			self.position[0],
			self.position[0] + self.length,
			self.position[1],
			self.position[1] + self.width,
			self.position[2],
			self.position[2] + self.height,
		])
		boundaries.shape = (3,2)
		return boundaries
	
	def rotate(self):
		'''
		Rotate the Object

		Rotations are treated as 90 degree rotations in the length x width
		plane.  Use the `is_rotated` attribute to determine if the object
		is in its original orientation or has been rotated.
		'''
		self.length, self.width = self.width, self.length
		self.length_raw, self.width_raw = self.width_raw, self.length_raw
		self.is_rotated = not self.is_rotated
	
	def get_weight(self,units=None):
		'''
		Get Weight of Logistics Object
		'''
		if units is None:
			return self.weight
		return VOLUME_CONVERTER[self.default_weight_units][units](self.weight)


class ShippingObject(LogisticsObject):
	'''
	Generalized Piece or Shipment that is Meant to be Moved Around Inside a Parent LogisticsObject
	'''

	def __init__(self,*args,**kwargs):
		super(ShippingObject,self).__init__(*args,**kwargs)
		self._binpack_item_exists = False
	
	def _set_binpack_item(self,**kwargs):
		'''
		Create a Bin Packing Object Associated to This Shipping Object
		'''
		if not self._binpack_item_exists:
			width = kwargs.get('width',self.width)
			height = kwargs.get('height',self.height)
			length = kwargs.get('length',self.length)
			weight = kwargs.get('weight',self.weight)
			self.binpack_item = Item(
				name = kwargs.get('name',getattr(self,'name','Unnamed Shipping Object')),
				width = width, 
				height = length,
				depth = height,
				weight = weight
			)
			self._binpack_item_exists = True
	
	def _delete_binpack_item(self):
		'''
		Delete Bin Packing Object
		'''
		if hasattr(self,'binpack_item'):
			delattr(self,'binpack_item')
		self._binpack_item_exists = False
	
	def _update_binpack_item(self,*args,**kwargs):
		'''
		Delete and Recreate Bin Packing Object
		'''
		self._delete_binpack_item()
		self._set_binpack_item(*args,**kwargs)
	
	def plot(self,*args,**kwargs):
		'''
		Plot the Container in the Length x Width Dimension
		'''
		boundaries = self.get_boundaries()
		x_list,y_list = get_rectangle_plot_lists(
			ll=kwargs.pop('h_offset',0) + boundaries[:,0], 
			ur=kwargs.pop('v_offset',0) + boundaries[:,1], 
			x_padding=kwargs.pop('h_padding',0), 
			y_padding=kwargs.pop('v_padding',0)
		)
		pyplot.plot(x_list,y_list,**kwargs)

	def fill(self,*args,**kwargs):
		'''
		Fill Plot the Container in the Length x Width Dimension
		'''
		boundaries = self.get_boundaries()
		x_list,y_list = get_rectangle_plot_lists(
			ll=kwargs.pop('h_offset',0) + boundaries[:,0], 
			ur=kwargs.pop('v_offset',0) + boundaries[:,1], 
			x_padding=kwargs.pop('h_padding',0), 
			y_padding=kwargs.pop('v_padding',0)
		)
		pyplot.fill(x_list,y_list,**kwargs)




from typing import Dict
from .trailer_load import optimize_trailer_load_plan
from ..exceptions import (
	TooManyPiecesException,
	NoPiecesException,
	PiecesTooLongForServiceException,
	OptimizationFailedServiceException,
	TrailerLoadingException,
	InvalidTrailerDimensionsException,
	InvalidPiecesException,
)
from ..optimizer_functions import PIECE_ARRANGEMENT_ROUTER, SHIPMENT_ARRANGEMENT_ROUTER
from ..standard_logistic_dims import STANDARD_TRAILER_DIMS
from .. import options
from copy import deepcopy
import numpy as np
import json

STANDARD_TRAILER_DIM_MAP = {
	obj.get('code') : obj for obj in STANDARD_TRAILER_DIMS
}

class NumpyArrayEncoder(json.JSONEncoder):
	'''
	Add serialization rules for numpy int, float, and boolean objects
	'''
	def default(self, obj):
		if isinstance(obj, np.int_):
			return int(obj)
		if isinstance(obj, np.float_):
			return float(obj)
		if isinstance(obj, np.bool_):
			return bool(obj)
		return json.JSONEncoder.default(self, obj)

def parse_request_data(request_data : Dict):
	'''
	Parse Request Data for Trailer Load Optimization API Serivce

	Parameters
	-----------
	request_data : Dict
		Dictionary of the form 
	
	Returns
	-----------
	status_code : int
		Status code of 4** for invalid request data, 200 for valid request data
	parsed_request_data : Dict
		Parsed request data, {} if status_code is 4**
	errors : Dict
		Error dictionary, {} if status_code is 200
	'''
	parsed_request_data = {}
	
	# Parse trailer dims/equipment code
	if isinstance(request_data.get('trailer_dims'),dict):
		parsed_request_data.update({'trailer_dims' : request_data['trailer_dims']})
	elif request_data.get('equipment_code') in STANDARD_TRAILER_DIM_MAP:
		# map equipment_code to appropriate trailer_dims
		parsed_request_data.update({'trailer_dims' : STANDARD_TRAILER_DIM_MAP[request_data['equipment_code']]})
	elif isinstance(request_data.get('equipment_code'),str):
		# if equipment code is passed, but not an option return a 400 error
		return 400,{},{
			'error_code' : 'InvalidTrailerDimensionsException',
			'error_message' : 'Equipment code not recognized',
		}
	else:
		return 400,{},{
			'error_code' : 'InvalidTrailerDimensionsException',
			'error_message' : 'Invalid trailer dimensions or equipment type provided',
		}
	
	# Parse shipment list
	if isinstance(request_data.get('shipment_list'),list):
		parsed_request_data.update({'shipment_list' : request_data['shipment_list']})
	elif hasattr(request_data.get('shipment_list'),'__iter__'):
		parsed_request_data.update({'shipment_list' : list(request_data['shipment_list'])})
	else:
		return 400,{},{
			'error_code' : 'InvalidPiecesException',
			'error_message' : 'Invalid pieces provided',
		}
	if not all([isinstance(shipment,dict) for shipment in parsed_request_data['shipment_list']]):
		return 400,{},{
			'error_code' : 'InvalidPiecesException',
			'error_message' : 'Invalid pieces provided',
		}
	for shipment in parsed_request_data['shipment_list']:
		shipment.update({
			"dimension_unit_of_measure": options.DimUomInches,
			"weight_unit_of_measure": options.WeightUomPounds,
		})

	# Parse allow rotations parameter
	if request_data.get('allow_rotations') == False:
		parsed_request_data.update({'allow_rotations' : False})
	elif request_data.get('allow_rotations') == True or request_data.get('allow_rotations') is None:
		# if allow_rotation is passed as True or not provided, set true
		parsed_request_data.update({'allow_rotations' : True})
	else:
		# If allow rotations is not None and non-boolean, raise validation error
		return 400,{},{
			'error_code' : 'InvalidPiecesException',
			'error_message' : 'Invalid allow rotations parameter',
		}
	
	if request_data.get('overweight_shipment_threshold') is not None:
		try:
			parsed_request_data.update({'overweight_shipment_threshold' : float(request_data['overweight_shipment_threshold'])})
		except:
			# If allow rotations is not None and non-boolean, raise validation error
			return 400,{},{
				'error_code' : 'InvalidPiecesException',
				'error_message' : 'Invalid overweight shipment threshold parameter',
			}
	
	if request_data.get('piece_arrangement_algorithm') is not None:
		if request_data.get('piece_arrangement_algorithm') in PIECE_ARRANGEMENT_ROUTER:
			parsed_request_data.update({'piece_arrangement_algorithm' : request_data['piece_arrangement_algorithm']})
		else:
			return 400,{},{
				'error_code' : 'OptimizationFailedServiceException',
				'error_message' : 'Invalid piece arrangement parameter provided',
			}
	
	if request_data.get('shipment_optimization_ls') is not None:
		try:
			for params in request_data.get('shipment_optimization_ls'):
				assert isinstance(params,dict)
				assert params['algorithm'] in SHIPMENT_ARRANGEMENT_ROUTER
				assert isinstance(params.get('max_iter'),(int,type(None)))
				assert isinstance(params.get('timeout'),(int,float,type(None)))
			parsed_request_data.update({'shipment_optimization_ls' : request_data['shipment_optimization_ls']})
		except:
			return 400,{},{
				'error_code' : 'OptimizationFailedServiceException',
				'error_message' : 'Invalid shipment arrangement parameter provided',
			}

	return 200,parsed_request_data,{}


def optimize_trailer_load_plan_wrapper(request_data : Dict):
	'''
	Trailer Load Optimization Function Intended for Use with API

	Parameters
	------------
	request_data : Dict
		Request data for trailer loading optimization
	
	Returns
	------------
	status_code : int 
		Status code to be used for API response
	response_dict : Dict
		Response, trailer loading result when status_code is 200, error summary when status_code is not 200
	'''
	# Parse request data
	try:
		request_status_code,parsed_request_data,errors = parse_request_data(request_data=deepcopy(request_data))
		if request_status_code != 200:
			return request_status_code,errors
	except:
		return 400,{
			'error_code' : 'InvalidRequestException',
			'error_message' : 'Invalid request',
			'request' : request_data,
		}
	
	# Do trailer load optimization
	try:
		trailer = optimize_trailer_load_plan(**parsed_request_data)
		if not trailer.arrangement_is_valid():
			return 500,{
				'error_code' : 'TrailerLoadingException',
				'error_message' : 'Unknown error',
			}
		trailer_load_plan = trailer.get_summary()
		trailer_load_plan = json.loads(json.dumps(trailer_load_plan,cls=NumpyArrayEncoder))
		return 200,trailer_load_plan
	except TooManyPiecesException:
		return 400,{
			'error_code' : 'TooManyPiecesException',
			'error_message' : 'Too many pieces provided',
		}
	except NoPiecesException:
		return 400,{
			'error_code' : 'NoPiecesException',
			'error_message' : 'No pieces provided',
		}
	except PiecesTooLongForServiceException:
		return 400,{
			'error_code' : 'PiecesTooLongForServiceException',
			'error_message' : 'At least one piece is too long for equipment',
		}
	except InvalidTrailerDimensionsException:
		return 400,{
			'error_code' : 'InvalidTrailerDimensionsException',
			'error_message' : 'Invalid trailer dimensions provided',
		}
	except InvalidPiecesException:
		return 400,{
			'error_code' : 'InvalidPiecesException',
			'error_message' : 'One or more pieces have invalid dimensions',
		}
	except OptimizationFailedServiceException:
		return 500,{
			'error_code' : 'OptimizationFailedServiceException',
			'error_message' : 'Optimization failed',
		}
	except TrailerLoadingException:
		return 500,{
			'error_code' : 'TrailerLoadingException',
			'error_message' : 'Unknown error',
		}
	except Exception as e:
		return 500,{
			'error_code' : 'Exception',
			'error_message' : 'Unknown error',
		}
	



import time
import numpy as np

simple_movement_slide_back_shipment_arrangement_details = {
	'code' : 'SLIDE_BACK',
	'name' : 'Slide Back Shipment Arrangement',
	'desc' : 'Attempt to slide shipments back in the trailer where space is available - Cannot be run stand-alone, intended to be run after `NO_STACK_BIN_PACK`',
}

def slide_shipments_back(trailer, max_iter : int = None, timeout : float = None, **kwargs):
	'''
	Slide Shipment Back in Trailer

	Parameters
	------------
	trailer : logistics_objects.Trailer
		Trailer object for shipments to be arranged
	max_iter : int
		Number of iterations to run
	timeout : float 
		Timeout for main loop of optimization
	
	Returns
	------------
	None
	'''
	max_iter = max_iter or 5
	start_time = time.perf_counter()
	break_loop = False
	index_order = np.argsort([
        shipment.position[0] for shipment in trailer.shipments
    ])
	for count in range(max_iter):
		for i in index_order:
			shipment = trailer.shipments[i]
			directions = np.random.choice(a=['left', 'right'], replace=False, size=2)
			shipment.slide_back(trailer)
			for d in directions:
				if d == 'left':
					shipment.slide_left(trailer)
				else:
					shipment.slide_right(trailer)
				shipment.slide_back(trailer)
			
			if timeout is not None:
				if time.perf_counter() - start_time > timeout and count > 0:
					break_loop = True
					break
		if break_loop:
			break
	shipment._set_dims()
	trailer.balance()



from typing import List, Dict
import numpy as np

def validate_pieces(shipment_list : List[Dict]):
    '''
    Validate Provided List of Pieces is Valid

    Parameters
    ------------
    shipment_list : List[Dict]
        List of dictionaries describing piece dimensions
        Valid piece dictionaries require positive numeric values for keys `length`, `width`, `height`, and `weight`.

    Returns
    ------------
    None

    Raises
    ------------
    ValueError
        Raised when non-numeric, infinite, or NaN values are present for requried keys.
    '''
    for piece in shipment_list:
        for c in ['length', 'width', 'height', 'weight']:
            try:
                float(piece.get(c)) 
            except:
                raise ValueError(f'Invalid {c} for at least one pieces - Numeric values required')
            if float(piece.get(c))  <= 0:
                raise ValueError(f'Invalid {c} for at least one pieces - Positive values required')
            if piece.get(c) == np.inf:
                raise ValueError(f'Invalid {c} for at least one pieces - Infinity not allowed')
            if piece.get(c) is np.nan:
                raise ValueError(f'Invalid {c} for at least one pieces - NaN not allowed')


def validate_trailer_dims(trailer_dims : Dict):
    '''
    Validate Provided Trailer Dimensions are Valid

    Parameters
    ------------
    trailer_dims : Dict
        Dictionary of trailer dimension values
        Valid trailer dimensions require positive values for keys `inner_width`, inner_length`, `inner_height`, and `max_weight`

    Returns
    ------------
    None

    Raises
    ------------
    ValueError
        Raised when non-numeric, infinite, or NaN values are present for requried keys.
    '''
    for c in ['inner_width', 'inner_length', 'inner_height', 'max_weight']:
        try:
            float(trailer_dims.get(c))
        except:
            raise ValueError(f'Invalid {c} for trailer dimensions - Numeric values required')
        if float(trailer_dims.get(c)) <= 0:
            raise ValueError(f'Invalid {c} for trailer dimensions - Positive values required')
        if trailer_dims.get(c) == np.inf:
            raise ValueError(f'Invalid {c} for trailer dimensions - Infinity not allowed')
        if trailer_dims.get(c) is np.nan:
            raise ValueError(f'Invalid {c} for trailer dimensions - NaN not allowed')


def validate_piece_length(shipment_list : List[Dict], trailer_dims : Dict, allow_rotations : bool):
    '''
    Validate Provided Trailer Dimensions are Valid

    Parameters
    ------------
    shipment_list : List[Dict]
        List of dictionaries describing piece dimensions
        Checks that each provided piece individually fits on the trailer
    trailer_dims : Dict
        Dictionary of trailer dimension values

    Returns
    ------------
    None

    Raises
    ------------
    AssertionError
        Raised when piece and trailer dimension combination is invalid
    '''
    if allow_rotations:
        trailer_dim_ls = sorted([
            trailer_dims.get('inner_length'),
            trailer_dims.get('inner_width'),
        ]) + [trailer_dims.get('inner_height')]
    else:
        trailer_dim_ls = [
            trailer_dims.get('inner_length'),
            trailer_dims.get('inner_width'),
            trailer_dims.get('inner_height'),
        ]
    for piece in shipment_list:
        if allow_rotations:
            dims = sorted([
                piece.get('length'),
                piece.get('width'),
            ]) + [piece.get('height')]
        else:
            dims = [
                piece.get('length'),
                piece.get('width'),
                piece.get('height'),
            ]
        assert(all([x <= y for x, y in zip(dims, trailer_dim_ls)]))




from .piece import Piece
from .shipment import Shipment
from .trailer import Trailer

'''
OOP inheritance structure

LogisticsObject
	ShippingObject(LogisticsObject)
		Piece(ShippingObject) - Non-divisible items
		Shipment(ShippingObject) - Collections/stacks of Piece objects, treated as single items to be arranged in Trailer
	Trailer(LogisticsObject)
'''





__version__ = '1.0.0'

from .services.trailer_load import optimize_trailer_load_plan
from .services.trailer_load_api import optimize_trailer_load_plan_wrapper
from .standard_logistic_dims import STANDARD_TRAILER_DIMS
from .exceptions import (
    TooManyPiecesException,
    NoPiecesException,
    PiecesTooLongForServiceException,
    OptimizationFailedServiceException,
    TrailerLoadingException,
    InvalidTrailerDimensionsException,
    InvalidPiecesException,
)




class TooManyPiecesException(Exception):
    pass


class NoPiecesException(Exception):
    pass


class PiecesTooLongForServiceException(Exception):
    pass


class OptimizationFailedServiceException(Exception):
    pass


class TrailerLoadingException(Exception):
    pass


class InvalidTrailerDimensionsException(Exception):
    pass


class InvalidPiecesException(Exception):
    pass


class InvalidOptimizationResultException(Exception):
    pass




from typing import List
from ...utils import (
	get_current_trailer_configuration,
	restore_trailer_configuration,
)
import time
import numpy as np

greedy_shipment_arrangement_details = {
	'code' : 'GREEDY_LOAD',
	'name' : 'Greedy Shipment Arrangement',
	'desc' : 'A stochastic greedy optimization for efficiently loading shipments onto the trailer',
}

def loss_function(trailer, shipment=None, **kwargs):
	'''
	Loss Function Intended to Score Trailer Load Plan Based on Trailer Linear Feet
	'''
	if trailer.arrangement_is_valid(shipment=shipment):
		return max([s.position[0] + s.length for s in trailer.shipments])
	return np.Infinity

def load_shipment(trailer, shipment, allow_rotations=True):
	'''
	Load Shipment on Trailer
	'''
	ll_candidate_positions = []
	ll_candidate_positions += [
		s.position + np.array([0, s.width, 0])
			for s in trailer.shipments[-5:] if s.position[1] + shipment.width < trailer.width
	]
	ll_candidate_positions += [
		s.position + np.array([s.length, 0, 0])
			for s in trailer.shipments[-5:] if s.position[1] + shipment.width < trailer.width
	]
	ll_candidate_positions += [trailer.position + np.array([0, 0, 0])]
	ul_candidate_positions = []
	ul_candidate_positions += [
		s.position + np.array([s.length, s.width, 0])
			for s in trailer.shipments[-5:] if s.position[1] - shipment.width > 0
	]
	ul_candidate_positions += [
		s.position + np.array([0, 0, 0]) 
			for s in trailer.shipments[-5:] if s.position[1] - shipment.width > 0
	]
	ul_candidate_positions += [trailer.position + np.array([0, trailer.width, 0])]
	trailer.add_shipment(shipment)
	position_loss = []
	for ll_candidate in ll_candidate_positions:
		shipment.position = ll_candidate
		loss = loss_function(trailer, shipment)
		position_loss += [
			(
				ll_candidate, 
				loss, 
				False
			)
		]
	for ul_candidate in ul_candidate_positions:
		shipment.position = ul_candidate - np.array([0, shipment.width, 0])
		loss = loss_function(trailer, shipment)
		position_loss += [
			(
				ul_candidate - np.array([0, shipment.width, 0]), 
				loss, 
				False
			)
		]
	if allow_rotations:
		shipment.rotate()
		for ll_candidate in ll_candidate_positions:
			shipment.position = ll_candidate
			loss = loss_function(trailer, shipment)
			position_loss += [
				(
					ll_candidate, 
					loss, 
					True
				)
			]
		for ul_candidate in ul_candidate_positions:
			shipment.position = ul_candidate - np.array([0, shipment.width, 0])
			loss = loss_function(trailer, shipment)
			position_loss += [
				(
					ul_candidate - np.array([0, shipment.width, 0]), 
					loss, 
					True
				)
			]
	if all([loss == np.Infinity for pos, loss, rot in position_loss]):
		if allow_rotations:
			shipment.rotate()
		shipment.position = np.array([trailer.get_boundaries()[0, 1], 0, 0])
		return loss_function(trailer, shipment)
	min_idx = np.argmin([loss for pos, loss, rot in position_loss])
	shipment.position, loss, rot = position_loss[min_idx]
	if allow_rotations and (not rot):
		shipment.rotate()
	shipment.slide_back(trailer)
	return loss_function(trailer, shipment)

def tetris_trailer_load(trailer, index_order : List, allow_rotations : bool = True, random : bool = True):
	'''
	Load Trailer with Tetris Style Placement with Random Order or Size Descending Order
	'''
	shipments = trailer.shipments
	for shipment in shipments:
		shipment.position = np.array([0, 0, 0])
	trailer.shipments = []
	for idx in index_order:
		shipment = shipments[idx]
		loss = load_shipment(
			trailer=trailer,
			shipment=shipment,
			allow_rotations=allow_rotations,
		)
	return loss


def get_sort_attr(s,r):
	'''
	Fuzzy Size Descending Sort Attribute for Shipments
	'''
	if r == 0:
		return (s.length * s.width) ** .5
	elif r == 1:
		return s.length
	elif r == 2:
		return s.width

def greedy_trailer_load(trailer, max_iter : int = None, timeout : float = None, allow_rotations : bool=True, **kwargs):
	'''
	Arrange Shipments in Trailer Using a Greedy Loading Algorithm
	
	Parameters
	------------
	trailer : logistics_objects.Trailer
		Trailer to be loaded
	max_iter : int
		Maximum number of iterations to be used for the optimization
	timeout : float 
		Timeout for main loop of optimization
	allow_rotations : bool
		Allow shipments to be rotated when loading
	
	Returns
	------------
	None
	'''
	max_iter = max_iter or 3
	temp_trailer_length = trailer.length
	trailer.length = sum([s.length for s in trailer.shipments])
	min_loss = np.Infinity
	shipment_arrangement = None
	start_time = time.perf_counter()
	for _ in range(max_iter):
		random_ls = np.random.choice(a=[0,1,2],size=len(trailer.shipments),replace=True)
		prob = np.array([get_sort_attr(s,r) for s,r in zip(trailer.shipments,random_ls)])
		prob = prob / np.sum(prob)
		index_order = [
			int(x) for x in np.random.choice(
				a=len(trailer.shipments),
				p=prob,
				size=len(trailer.shipments),
				replace=False
			)
		]
		loss = tetris_trailer_load(
			trailer=trailer,
			allow_rotations=allow_rotations,
			index_order = index_order,
		)
		if loss < min_loss:
			shipment_arrangement = get_current_trailer_configuration(trailer)
			min_loss = loss
		if timeout is not None:
			if time.perf_counter() - start_time > timeout:
				break
	restore_trailer_configuration(shipment_arrangement)
	trailer.length = temp_trailer_length
	for shipment in trailer.shipments:
		shipment._set_dims()




from .constants import (
    RotationType, 
    NoRotationType, 
    Axis,
)
from .auxiliary_methods import intersect, set_to_decimal

DEFAULT_NUMBER_OF_DECIMALS = 3
START_POSITION = [0, 0, 0]


class Item:
    def __init__(self, name, width, height, depth, weight):
        self.name = name
        self.width = width
        self.height = height
        self.depth = depth
        self.weight = weight
        self.rotation_type = 0
        self.position = START_POSITION
        self.number_of_decimals = DEFAULT_NUMBER_OF_DECIMALS

    def format_numbers(self, number_of_decimals):
        self.width = set_to_decimal(self.width, number_of_decimals)
        self.height = set_to_decimal(self.height, number_of_decimals)
        self.depth = set_to_decimal(self.depth, number_of_decimals)
        self.weight = set_to_decimal(self.weight, number_of_decimals)
        self.number_of_decimals = number_of_decimals

    def string(self):
        return "%s(%sx%sx%s, weight: %s) pos(%s) rt(%s) vol(%s)" % (
            self.name, self.width, self.height, self.depth, self.weight,
            self.position, self.rotation_type, self.get_volume()
        )

    def get_volume(self):
        return set_to_decimal(
            self.width * self.height * self.depth, self.number_of_decimals
        )

    def get_dimension(self):
        if self.rotation_type == RotationType.RT_WHD:
            dimension = [self.width, self.height, self.depth]
        elif self.rotation_type == RotationType.RT_HWD:
            dimension = [self.height, self.width, self.depth]
        elif self.rotation_type == RotationType.RT_HDW:
            dimension = [self.height, self.depth, self.width]
        elif self.rotation_type == RotationType.RT_DHW:
            dimension = [self.depth, self.height, self.width]
        elif self.rotation_type == RotationType.RT_DWH:
            dimension = [self.depth, self.width, self.height]
        elif self.rotation_type == RotationType.RT_WDH:
            dimension = [self.width, self.depth, self.height]
        else:
            dimension = []

        return dimension


class Bin:
    def __init__(self, name, width, height, depth, max_weight):
        self.name = name
        self.width = width
        self.height = height
        self.depth = depth
        self.max_weight = max_weight
        self.items = []
        self.unfitted_items = []
        self.number_of_decimals = DEFAULT_NUMBER_OF_DECIMALS

    def format_numbers(self, number_of_decimals):
        self.width = set_to_decimal(self.width, number_of_decimals)
        self.height = set_to_decimal(self.height, number_of_decimals)
        self.depth = set_to_decimal(self.depth, number_of_decimals)
        self.max_weight = set_to_decimal(self.max_weight, number_of_decimals)
        self.number_of_decimals = number_of_decimals

    def string(self):
        return "%s(%sx%sx%s, max_weight:%s) vol(%s)" % (
            self.name, self.width, self.height, self.depth, self.max_weight,
            self.get_volume()
        )

    def get_volume(self):
        return set_to_decimal(
            self.width * self.height * self.depth, self.number_of_decimals
        )

    def get_total_weight(self):
        total_weight = 0

        for item in self.items:
            total_weight += item.weight

        return set_to_decimal(total_weight, self.number_of_decimals)

    def put_item(self, item, pivot):
        fit = False
        valid_item_position = item.position
        item.position = pivot

        for i in range(0, len(RotationType.ALL)):
            item.rotation_type = i
            dimension = item.get_dimension()
            if (
                self.width < pivot[0] + dimension[0] or
                self.height < pivot[1] + dimension[1] or
                self.depth < pivot[2] + dimension[2]
            ):
                continue

            fit = True

            for current_item_in_bin in self.items:
                if intersect(current_item_in_bin, item):
                    fit = False
                    break

            if fit:
                if self.get_total_weight() + item.weight > self.max_weight:
                    fit = False
                    return fit

                self.items.append(item)

            if not fit:
                item.position = valid_item_position

            return fit

        if not fit:
            item.position = valid_item_position

        return fit


class NoRotateBin(Bin):
    def put_item(self, item, pivot):
        fit = False
        valid_item_position = item.position
        item.position = pivot

        for i in range(0, len(NoRotationType.ALL)):
            item.rotation_type = i
            dimension = item.get_dimension()
            if (
                self.width < pivot[0] + dimension[0] or
                self.height < pivot[1] + dimension[1] or
                self.depth < pivot[2] + dimension[2]
            ):
                continue

            fit = True

            for current_item_in_bin in self.items:
                if intersect(current_item_in_bin, item):
                    fit = False
                    break

            if fit:
                if self.get_total_weight() + item.weight > self.max_weight:
                    fit = False
                    return fit

                self.items.append(item)

            if not fit:
                item.position = valid_item_position

            return fit

        if not fit:
            item.position = valid_item_position

        return fit


class Packer:
    def __init__(self):
        self.bins = []
        self.items = []
        self.unfit_items = []
        self.total_items = 0

    def add_bin(self, bin):
        return self.bins.append(bin)

    def add_item(self, item):
        self.total_items = len(self.items) + 1

        return self.items.append(item)

    def pack_to_bin(self, bin, item):
        fitted = False

        if not bin.items:
            response = bin.put_item(item, START_POSITION)

            if not response:
                bin.unfitted_items.append(item)

            return

        for axis in range(0, 3):
            items_in_bin = bin.items

            for ib in items_in_bin:
                pivot = [0, 0, 0]
                w, h, d = ib.get_dimension()
                if axis == Axis.WIDTH:
                    pivot = [
                        ib.position[0] + w,
                        ib.position[1],
                        ib.position[2]
                    ]
                elif axis == Axis.HEIGHT:
                    pivot = [
                        ib.position[0],
                        ib.position[1] + h,
                        ib.position[2]
                    ]
                elif axis == Axis.DEPTH:
                    pivot = [
                        ib.position[0],
                        ib.position[1],
                        ib.position[2] + d
                    ]

                if bin.put_item(item, pivot):
                    fitted = True
                    break
            if fitted:
                break

        if not fitted:
            bin.unfitted_items.append(item)

    def pack(
        self, bigger_first=False, distribute_items=False,
        number_of_decimals=DEFAULT_NUMBER_OF_DECIMALS
    ):
        for bin in self.bins:
            bin.format_numbers(number_of_decimals)

        for item in self.items:
            item.format_numbers(number_of_decimals)

        self.bins.sort(
            key=lambda bin: bin.get_volume(), reverse=bigger_first
        )
        self.items.sort(
            key=lambda item: item.get_volume(), reverse=bigger_first
        )

        for bin in self.bins:
            for item in self.items:
                self.pack_to_bin(bin, item)

            if distribute_items:
                for item in bin.items:
                    self.items.remove(item)




from ...py3dbp import Packer

no_stack_bin_pack_shipment_arrangement_details = {
	'code' : 'NO_STACK_BIN_PACK',
	'name' : 'No-Stacking Bin Pack Shipment Arrangement',
	'desc' : 'Bin packing algorithm with stacking dissallowed to load shipments onto the trailer',
}

def no_stack_pack_trailer(trailer, trailer_occupy_length : int, allow_rotations=True):
	'''
	Apply Bin Packing Algorithm to Load a Portion of a Trailer

	Parameters
	------------
	trailer : logistics_objects.Trailer
		Trailer object to attempt to load
	trailer_occupy_length : int
		Number of inches of the trailer to allow for loading
	
	Returns
	------------
	loaded_successfully : bool
		True if all items fit in the trailer, False if not
	'''
	trailer._update_binpack_bin(
		height=1.5,
		length=trailer_occupy_length,
		allow_rotations=allow_rotations,
	)

	packer = Packer()
	packer.add_bin(trailer.binpack_bin)
	for shipment in trailer.shipments:
		packer.add_item(shipment.binpack_item)

	packer.pack(
		distribute_items=False,
		bigger_first=True
	)
	b = packer.bins[0]
	return len(b.unfitted_items) == 0


def no_stack_trailer_arrange(trailer, allow_rotations=True,search_method : str = 'bisection', **kwargs):
	'''
	Allocate Shipments in Trailer using an Incremental Bin Packing Algorithm without Stacking

	Parameters
	------------
	trailer : logistics_objects.Trailer
		Trailer object to be optimally loaded
	allow_rotations : bool
		Allow shipments to be rotated in order to load
	search_method : str
		Method for searching for optimal portion of trailer, options: `bisection`, `foot_increment`

	Returns
	------------
	None
	'''

	for shipment in trailer.shipments:
		shipment.position[0] = 0
		shipment.position[1] = 0

	all_shipment_loaded = False
	for shipment in trailer.shipments:
		shipment._update_binpack_item(height=1)
	
	if search_method == 'foot_increment':
		trailer_occupy_length = min([shipment.length for shipment in trailer.shipments]) - 12
		for _ in range(100):
			trailer_occupy_length += 12
			all_shipment_loaded = no_stack_pack_trailer(
				trailer=trailer,
				trailer_occupy_length=trailer_occupy_length,
				allow_rotations=allow_rotations,
			)
			if all_shipment_loaded:
				break
	else:
		a = 0
		b = sum([shipment.length for shipment in trailer.shipments])
		c = round((a+b)/2)
		count = 0
		while count < 25:
			count += 1
			all_shipment_loaded = no_stack_pack_trailer(
				trailer=trailer,
				trailer_occupy_length=c,
				allow_rotations=allow_rotations,
			)
			if all_shipment_loaded:
				b = c
			else:
				a = c
			c = round((a+b)/2)
			if c in (a,b):
				break
		all_shipment_loaded = no_stack_pack_trailer(
			trailer=trailer,
			trailer_occupy_length=b,
			allow_rotations=allow_rotations,
		)
	for shipment in trailer.shipments:
		if shipment.binpack_item.rotation_type == 1:
			shipment.rotate()
		shipment.position[0] = float(shipment.binpack_item.position[1])
		shipment.position[1] = float(shipment.binpack_item.position[0])




from .common import ShippingObject
from ..options import PACKING_OPTIONS

class Piece(ShippingObject):
	'''
	Indivisible Box or Pallet

	This shipping object is intended to be an indivisible
	shipping item, i.e. one that must be loaded without
	being taken apart or separated.
	'''

	def __init__(self,packing,stack_limit,commodity=None,value=None,*args,**kwargs):
		super(Piece,self).__init__(*args,**kwargs)
		# Ensure a valid packing type is provided
		assert packing in list(map(lambda x : x[0],PACKING_OPTIONS))
		# Ensure a valid stack_limit is provided
		assert stack_limit > 0

		self.packing = packing
		self.commodity = commodity
		self.value = value
		self.stack_limit = stack_limit

		self._set_binpack_item()





from .common import ShippingObject
from .piece import Piece
from ..py3dbp import (
	Bin,
	NoRotateBin,
)
from ..utils import intervals_overlap
import numpy as np

class Shipment(ShippingObject):
	'''
	Shipment that May be made up of Multiple Pieces

	This is a shipment unit that may be composed of multiple pieces
	stacked or otherwise put together that are intended to be loaded
	and transported as a single unit.  It may be a single piece or 
	it may be many pieces put together.
	'''
	def __init__(self,pieces=[],*args,**kwargs):
		if isinstance(pieces,Piece):
			pieces = [pieces]
		self.pieces = []
		self.add_piece(pieces)
		self.arrange_pieces()

		super(Shipment,self).__init__(
			length = self.length,
			width = self.width,
			height = self.height,
			dimension_unit_of_measure = self.default_dimension_units,
			weight = self.weight,
			weight_unit_of_measure = self.default_weight_units,
			*args,
			**kwargs
		)
		self._binpack_bin_exists = False
		self.max_weight = 100000

	def arrange_pieces(self,**kwargs):
		'''
		Trivial Piece Arragemet:  Place Shipment End-to-End in the Length direction
		'''
		length_pos = 0
		for piece in self.pieces:
			piece.position[0] = length_pos
			length_pos += piece.length
		# Update shipment dimension/weight attributes
		self._set_dims()

	def _set_dims(self):
		'''
		Update Dimensions, Weight, and Piece Ordering for Shipment
		'''
		if len(self.pieces) == 0:
			self.length = 0
			self.width = 0
			self.height = 0
			self.weight = 0
		else:
			self.length = max([piece.position[0] + piece.length for piece in self.pieces]) - min([piece.position[0] for piece in self.pieces])
			self.width = max([piece.position[1] + piece.width for piece in self.pieces]) - min([piece.position[1] for piece in self.pieces])
			self.height = max([piece.position[2] + piece.height for piece in self.pieces]) - min([piece.position[2] for piece in self.pieces])
			self.weight = sum([piece.weight for piece in self.pieces])
		stack_order = np.argsort([p.position[2] for p in self.pieces])
		self.pieces = [self.pieces[i] for i in stack_order]

	def add_piece(self,piece):
		'''
		Add Piece to Shipment
		'''
		if isinstance(piece,Piece):
			piece = [piece]
		assert all([isinstance(p,Piece) for p in piece])
		self.pieces += piece

	def remove_piece(self,piece):
		'''
		Remove Piece from Shipment
		'''
		if isinstance(piece,Piece):
			self.pieces = [p for p in self.pieces if p != piece]
			return piece
		elif isinstance(piece,int):
			return self.pieces.pop(piece)
		else:
			raise Exception(f'Data type for {piece} not recognized')

	def get_boundaries(self):
		'''
		Get Shipment Profile
		'''
		boundaries = np.array([np.inf,-np.inf,np.inf,-np.inf,np.inf,-np.inf])
		boundaries.shape = (3,2)
		for piece in self.pieces:
			piece_boundaries = piece.get_boundaries()
			for i in range(3):
				boundaries[i,0] = min([boundaries[i,0],piece_boundaries[i,0]])
				boundaries[i,1] = max([boundaries[i,1],piece_boundaries[i,1]])
		for i in range(3):
			boundaries[i,0] += self.position[i]
			boundaries[i,1] += self.position[i]
		return boundaries
	
	def _set_binpack_bin(self,allow_rotations=True,**kwargs):
		'''
		Set Bin Pack Items
		'''
		if not self._binpack_bin_exists:
			width = kwargs.get('width',self.width)
			height = kwargs.get('height',self.height)
			length = kwargs.get('length',self.length)
			max_weight = kwargs.get('max_weight',self.weight)
			if allow_rotations:
				self.binpack_bin = Bin(
					name = kwargs.get('name',getattr(self,'name','Unnamed Shipment Object')),
					width = width, 
					height = length,
					depth = height,
					max_weight = max_weight
				)
			else:
				self.binpack_bin = NoRotateBin(
					name = kwargs.get('name',getattr(self,'name','Unnamed Shipment Object')),
					width = width, 
					height = length,
					depth = height,
					max_weight = max_weight
				)
			self._binpack_bin_exists = True
	
	def _delete_binpack_bin(self):
		'''
		Delete Bin Pack Item
		'''
		if hasattr(self,'binpack_item'):
			delattr(self,'binpack_item')
		self._binpack_bin_exists = False
	
	def _update_binpack_bin(self,allow_rotations=True,*args,**kwargs):
		'''
		Delete and Recreate Bin Pack Item
		'''
		self._delete_binpack_bin()
		self._set_binpack_bin(allow_rotations=allow_rotations,*args,**kwargs)
	
	def arrangement_is_valid(self):
		'''
		Validate Positions of Shipment's Pieces are Physically Valid
		'''
		piece_boundaries_list = [piece.get_boundaries() for piece in self.pieces]
		for i in range(len(piece_boundaries_list)):
			p_i = piece_boundaries_list[i]
			for j in range(i):
				p_j = piece_boundaries_list[j]
				overlap = all([intervals_overlap(p_i[k],p_j[k]) for k in range(3)])
				if overlap:
					return False
		return True

	def rotate(self):
		self.length, self.width = self.width, self.length
		self.length_raw, self.width_raw = self.width_raw, self.length_raw
		self.is_rotated = not self.is_rotated
		
		# Also rotate all pieces within the shipment
		for piece in self.pieces:
			piece.position[0],piece.position[1] = piece.position[1],piece.position[0]
			piece.length, piece.width = piece.width, piece.length
			piece.length_raw, piece.width_raw = piece.width_raw, piece.length_raw
			piece.is_rotated = not piece.is_rotated
	
	def move(self,delta):
		'''
		Move the Shipment
		'''
		self.position += delta
	
	def slide_back(self,trailer):
		'''
		Slide Shipment Back in Length Direction as far as it can Within the Trailer
		'''
		delta = np.array([-1,0,0])
		count = 0
		while trailer.arrangement_is_valid(self) and count < 1000:
			self.move(delta)
			count += 1
		self.move(-delta)

	def slide_right(self,trailer):
		'''
		Slide Shipment to the Right in the Width Direction as far as it can Within the Trailer
		'''
		delta = np.array([0,1,0])
		count = 0
		while trailer.arrangement_is_valid(self) and count < 1000:
			self.move(delta)
			count += 1
		self.move(-delta)

	def slide_left(self,trailer):
		'''
		Slide Shipment to the Left in the Width Direction as far as it can Within the Trailer
		'''
		delta = np.array([0,-1,0])
		count = 0
		while trailer.arrangement_is_valid(self) and count < 1000:
			self.move(delta)
			count += 1
		self.move(-delta)

	def can_stack(self,piece,trailer_height):
		'''
		Check if Provided Piece can be Stacked on top of the Shipment
		'''
		# TODO: Add other stacking options - stack next to on top of
		# TODO: Add maximum weight/pressure restrictions for pieces 
		min_length = min([p.length for p in self.pieces])
		min_width = min([p.width for p in self.pieces])
		
		total_hieght = sum([p.height for p in self.pieces]) + piece.height
		
		current_stack = len(self.pieces)
		stack_limit = min([p.stack_limit for p in self.pieces] + [piece.stack_limit])
		
		# Allow stacking pallets on pallets, boxes on boxes, and boxes on pallets, but not pallets on boxes.
		packing_stackable = not (
			piece.packing == 'PALLET' and 
			any([p.packing == 'BOX' for p in self.pieces])
		)
		return (
			(piece.length <= min_length) and 
			(piece.width <= min_width) and 
			(total_hieght <= trailer_height) and 
			(current_stack < stack_limit) and 
			packing_stackable
		)
		
	def stack(self,piece):
		'''
		Stack Piece on Shipment
		'''
		# Allow stacking in positions other than 0,0 in the length x width dimension
		self._set_dims()
		bottom_piece = self.pieces[-1]
		piece.position = bottom_piece.position + np.array([0,0,bottom_piece.height])
		self.add_piece(piece)
		self._set_dims()




PackingUnitPallet = 'PALLET'
PackingUnitBox = 'BOX'

PACKING_OPTIONS = [
    (PackingUnitPallet, 'Palletized Freight'),
    (PackingUnitBox, 'Box'),
]

DimUomInches = 'IN'
DimUomFeet = 'FT'
DimUomMeter = 'METER'

DIMENSION_UNIT_OF_MEASURE_OPTIONS = [
    (DimUomInches, 'Inches'),
    (DimUomFeet, 'Feet'),
    (DimUomMeter, 'Meters'),
]

VolumeUomCubicInches = 'CUBIC_IN'
VolumeUomCubicFeet = 'CUBIC_FT'
VolumeUomCubicMeters = 'CUBIC_METER'

VOLUME_UNIT_OF_MEASURE_OPTIONS = [
    (VolumeUomCubicInches, 'Cubic Inches'),
    (VolumeUomCubicFeet, 'Cubic Feet'),
    (VolumeUomCubicMeters, 'Cubic Meters'),
]

WeightUomPounds = 'LBS'
WeightUomKilgrams = 'KG'
WeightUomGrams = 'G'

WEIGHT_UNIT_OF_MEASURE_OPTIONS = [
    (WeightUomPounds, 'Pound'),
    (WeightUomKilgrams, 'Kilograms'),
    (WeightUomGrams, 'Grams'),
]




from typing import List
from ...logistics_objects import Shipment

naive_piece_arrangement_details = {
	'code' : 'NAIVE',
	'name' : 'Naive Piece Arrangement',
	'desc' : 'The trivial piece packing - make each piece its own shipment with no stacking',
}

def naive_piece_arrangement(pieces : List, **kwargs):
	'''
	Allocate Provided Pieces into Shipments in a Trivia/Naive Manner

	No optimization is done in this piece-shipment allocation, just 
	one shipment per piece with no stacking.

	Parameters
	------------
	pieces : List[logistics_objects.Piece]
		List of pieces to be allocated into shipments
	
	Returns
	-----------
	shipments : List[logistics_objects.Shipment]
		List of shipments composed of provided piece objects
	'''
	shipments = [Shipment(pieces=piece) for piece in pieces]
	for shipment in shipments:
		shipment._set_dims()
	return shipments




from typing import List, Dict
import numpy as np


def intervals_overlap(a : List[float], b : List[float]):
	'''
	Determine Two Provided Intervals Overlap
	
	Params
	-----------
	a : List[float]
		List, or list-like object, with numeric values of length 2
	b : List[float]
		List, or list-like object, with numeric values of length 2
	
	Returns
	-----------
	overlap : bool
		True if intervals a and b overlap, False if not
	
	Examples
	-----------
	intervals_overlap([0,1],[1,2]) # don't overlap, return False
	intervals_overlap([1,2],[0,1]) # don't overlap, return False
	intervals_overlap([0,2],[1,3]) # overlap, return True
	intervals_overlap([1,3],[0,2]) # overlap, return True
	intervals_overlap([0,5],[1,2]) # overlap, return True
	intervals_overlap([1,2],[0,5]) # overlap, return True
	'''
	if a[0] < b[1]:
		return a[1] > b[0]
	else:
		return a[0] < b[1]


def interval_is_subset(a : List[float], b : List[float]):
	'''
	Determine First Interval is a Subset of the Second
	
	Params
	-----------
	a : List
		List, or list-like object, with numeric values of length 2
	b : List
		List, or list-like object, with numeric values of length 2
	
	Returns
	-----------
	overlap : bool
		True if interval a is a subset of b, False if not
	
	Examples
	-----------
	interval_is_subset([1,2],[0,3]) # returns True
	interval_is_subset([0,2],[0,3]) # returns True
	interval_is_subset([0,2],[1,3]) # returns False
	interval_is_subset([1,4],[0,3]) # returns False
	'''
	return a[0] >= b[0] and a[1] <= b[1]


def rotation_matrix(theta : float):
	'''
	Construct a 3x3 Rotation Matrix Rotating Counter-Clockwise in the First Two Dimensions

	Parameters
	--------------
	theta : float
		Degrees of rotation in radians
	
	Returns
	--------------
	rot_matrix
		3 x 3 rotation matrix - rotation of theta degrees counter-clockwise in the x-y plane (orthogonal to the third dimension)
	'''
	return np.array([
		[np.cos(theta), -np.sin(theta), 0],
		[np.sin(theta), np.cos(theta), 0],
		[0, 0, 1],
	])


def list_remove(idx : int, allocated : List, available : List):
	'''
	Remove Element from List While Managing Available/Unavailable Selections

	Params
	----------
	idx : int
		Index of value to be removed
	
	Returns
	----------
	None
	'''
	allocated.remove(idx)
	available += [idx]


def list_allocate(idx : int, allocated : List, available : List):
	'''
	Allocate Element from List While Managing Available/Unavailable Selections

	Params
	----------
	idx : int
		Index of value to be allocated
	
	Returns
	----------
	None
	'''
	available.remove(idx)
	allocated += [idx]

def get_current_trailer_configuration(trailer):
	'''
	Get Piece Location Details for Current Trailer Configuration
	
	Get the details of the current configuration for a trailer object
	to be used as a backup to be restored at a later time.  To be 
	used in conjunction with `restore_trailer_configuration`.

	Params
	---------
	trailer : logistics_objects.Trailer
		Trailer object for which to save current configuration
	
	Returns
	---------
	shipment_arrangement : List[Dict]
		List of dictionaries detailing the location of pieces in the current trailer load plan
	'''
	return [(s, s.position, s.is_rotated) for s in trailer.shipments]


def restore_trailer_configuration(shipment_arrangement : List[Dict]):
	'''
	Restore Provided Piece Location Configuration
	
	Restore a piece configuration inside a trailer.  To be used
	in conjuction with `get_current_trailer_configuration` to 
	restore a previous state of the trailer object.

	Params
	---------
	shipment_arrangement : List[Dict]
		Shipment arrangment list - typically, this would be a value returned by `get_current_trailer_configuration`
	
	Returns
	---------
	None
	'''
	for s, pos, rot in shipment_arrangement:
		s.position = pos
		if rot != s.is_rotated:
			s.rotate()
		s._set_dims()




from .shipment_arrangement.naive import naive_shipment_arrangement, naive_shipment_arrangement_details
from .shipment_arrangement.bin_packing import no_stack_trailer_arrange, no_stack_bin_pack_shipment_arrangement_details
from .shipment_arrangement.simple_movements import slide_shipments_back, simple_movement_slide_back_shipment_arrangement_details
from .shipment_arrangement.greedy import greedy_trailer_load, greedy_shipment_arrangement_details

from .piece_arrangement.naive import naive_piece_arrangement, naive_piece_arrangement_details
from .piece_arrangement.greedy import greedy_stack_pieces, greedy_stack_piece_arrangement_details


PIECE_ARRANGEMENT_ROUTER = {
	'NAIVE': (naive_piece_arrangement, naive_piece_arrangement_details),
	'GREEDY_STACK': (greedy_stack_pieces, greedy_stack_piece_arrangement_details),
}

SHIPMENT_ARRANGEMENT_ROUTER = {
	'NAIVE': (naive_shipment_arrangement, naive_shipment_arrangement_details),
	'NO_STACK_BIN_PACK': (no_stack_trailer_arrange, no_stack_bin_pack_shipment_arrangement_details),
	'SLIDE_BACK': (slide_shipments_back, simple_movement_slide_back_shipment_arrangement_details),
	'GREEDY_LOAD': (greedy_trailer_load, greedy_shipment_arrangement_details),
}

def optimize_pieces_arrangement(pieces,algorithm : str,**kwargs):
	'''
	Piece Arrangement Optimization Router Function

	Optimizes the arrangement of provided pieces in to shipments by 
	collecting and stacking pieces into shipments to be loaded into
	a trailer.  Router function leverages `PIECE_ARRANGEMENT_ROUTER`
	to allow for various algorithm options.  All functions in the 
	piece arrangement router require `pieces` and algorithm 
	parameters, and return a list of shipments of type 
	List[logistics_objects.Shipment].  See individual functions in piece 
	arrangement router for additional required parameters.

	Parameters
	-------------
	pieces : List[logistics_objects.Piece]
		List of pieces to be arranged into shipments
	algorithm : str
		Algorithm key, options are the keys of `PIECE_ARRANGEMENT_ROUTER`

	Returns
	-------------
	shipments : List[logistics_objects.Shipment]
		List of shipments made up of provided pieces

	'''
	if not algorithm in PIECE_ARRANGEMENT_ROUTER.keys():
		raise NotImplementedError(f'Algorithm `{algorithm}` has not been implemented')
	return PIECE_ARRANGEMENT_ROUTER[algorithm][0](pieces=pieces,**kwargs)


def optimize_shipment_arrangement(trailer,algorithm,**kwargs):
	'''
	Shipment Arrangement Optimization Router Function

	Optimizes the arrangement of shipment associated to the provided
	logistics_objects.Trailer object.  Router function leverages 
	`SHIPMENT_ARRANGEMENT_ROUTER` to allow for various algoirthm
	options.  All functions in the shipment arrangement router require
	`trailer` and `algorithm` parameters, update the shipment 
	arrangement stored in the trailer object, and return None.
	
	Parameters
	-------------
	trailer : logistics_objects.Trailer
		Trailer object with shipments to be optimally loaded
	algorithm : str
		Algorithm key, options are the keys of `SHIPMENT_ARRANGEMENT_ROUTER`

	Returns
	-------------
	None
	'''
	if not algorithm in SHIPMENT_ARRANGEMENT_ROUTER.keys():
		raise NotImplementedError(f'Algorithm `{algorithm}` has not been implemented')
	return SHIPMENT_ARRANGEMENT_ROUTER[algorithm][0](trailer=trailer,**kwargs)




from . import options

'''
Generic Unility Functions for Conversting between Various Units

Usage is:

    DIMENSION_CONVERTER[<convert from unit : str>][<convert to unit : str>](<from unit value : float>)
        >> <to unit value : float>
    DIMENSION_CONVERTER['IN']['FT'](12)
        >> 1
    Similar usage for VOLUME_CONVERTER and WEIGHT_CONVERTER
'''

INCHES_PER_FOOT = 12
INCHES_PER_METER = 39.3701

POUNDS_PER_KILOGRAM = 2.20462
POUNDS_PER_GRAM = POUNDS_PER_KILOGRAM / 1000

DIMENSION_CONVERTER = {
    options.DimUomInches: {
        options.DimUomInches: lambda x: x,
        options.DimUomFeet: lambda x: x / INCHES_PER_FOOT,
        options.DimUomMeter: lambda x: x / INCHES_PER_METER,
    },
    options.DimUomFeet: {
        options.DimUomInches: lambda x: INCHES_PER_FOOT * x,
        options.DimUomFeet: lambda x: x,
        options.DimUomMeter: lambda x: x * INCHES_PER_FOOT / INCHES_PER_METER,
    },
    options.DimUomMeter: {
        options.DimUomInches: lambda x: INCHES_PER_METER * x,
        options.DimUomFeet: lambda x: x * INCHES_PER_METER / INCHES_PER_FOOT,
        options.DimUomMeter: lambda x: x,
    },
}

VOLUME_CONVERTER = {
    options.VolumeUomCubicInches: {
        options.VolumeUomCubicInches: lambda x: x,
        options.VolumeUomCubicFeet: lambda x: x / INCHES_PER_FOOT ** 3,
        options.VolumeUomCubicMeters: lambda x: x / INCHES_PER_METER ** 3,
    },
    options.VolumeUomCubicFeet: {
        options.VolumeUomCubicInches: lambda x: INCHES_PER_FOOT ** 3 * x,
        options.VolumeUomCubicFeet: lambda x: x,
        options.VolumeUomCubicMeters: lambda x: x * INCHES_PER_FOOT ** 3 / INCHES_PER_METER ** 3,
    },
    options.VolumeUomCubicMeters: {
        options.VolumeUomCubicInches: lambda x: INCHES_PER_METER ** 3 * x,
        options.VolumeUomCubicFeet: lambda x: x * INCHES_PER_METER ** 3 / INCHES_PER_FOOT ** 3,
        options.VolumeUomCubicMeters: lambda x: x,
    },
}

WEIGHT_CONVERTER = {
    options.WeightUomPounds: {
        options.WeightUomPounds: lambda x: x,
        options.WeightUomKilgrams: lambda x: x / POUNDS_PER_KILOGRAM,
        options.WeightUomGrams: lambda x: x / POUNDS_PER_GRAM,
    },
    options.WeightUomKilgrams: {
        options.WeightUomPounds: lambda x: x * POUNDS_PER_KILOGRAM,
        options.WeightUomKilgrams: lambda x: x,
        options.WeightUomGrams: lambda x: x * POUNDS_PER_KILOGRAM / POUNDS_PER_GRAM,
    },
    options.WeightUomGrams: {
        options.WeightUomPounds: lambda x: x * POUNDS_PER_GRAM,
        options.WeightUomKilgrams: lambda x: x * POUNDS_PER_GRAM / POUNDS_PER_KILOGRAM,
        options.WeightUomGrams: lambda x: x,
    },
}




from .common import LogisticsObject, get_rectangle_plot_lists
from .shipment import Shipment
from ..py3dbp import (
	Bin,
	NoRotateBin,
)
from ..unit_converters import (
	DIMENSION_CONVERTER,VOLUME_CONVERTER,
)
from ..utils import intervals_overlap,interval_is_subset
from .. import options
import numpy as np
from matplotlib import pyplot

class Trailer(LogisticsObject):
	'''
	Generalized Trailer Object
	'''
	
	def __init__(self,max_weight,shipments=[],name='Unnamed Trailer',*args,**kwargs):
		if isinstance(shipments,Shipment):
			shipments = [shipments]
		super(Trailer,self).__init__(*args,**kwargs)

		self.max_weight = max_weight
		self.name = name
		self.shipments = []
		self.add_shipment(shipments)

		self._binpack_bin_exists = False
		self.set_var_types(variables='max_weight',dtype=float)

	def get_boundaries(self):
		'''
		Get Trailer Profile
		'''
		if len(self.shipments) == 0:
			boundaries = np.array([0,0,0,0,0,0])
			boundaries.shape = (3,2)
			return boundaries
		boundaries = np.array([np.inf,-np.inf,np.inf,-np.inf,np.inf,-np.inf])
		boundaries.shape = (3,2)
		for shipment in self.shipments:
			shipment_boundaries = shipment.get_boundaries()
			for i in range(3):
				boundaries[i,0] = min([boundaries[i,0],shipment_boundaries[i,0]])
				boundaries[i,1] = max([boundaries[i,1],shipment_boundaries[i,1]])
		for i in range(3):
			boundaries[i,0] += self.position[i]
			boundaries[i,1] += self.position[i]
		return boundaries
	
	def add_shipment(self,shipment):
		'''
		Add Shipment to Trailer
		'''
		if isinstance(shipment,Shipment):
			shipment = [shipment]
		assert all([isinstance(s,Shipment) for s in shipment])
		self.shipments += shipment

	def remove_shipment(self,shipment):
		'''
		Remove Shipment from Trailer
		'''
		if isinstance(shipment,Shipment):
			self.shipments = [s for s in self.shipments if s != shipment]
			return shipment
		elif isinstance(shipment,int):
			return self.shipments.pop(shipment)
		else:
			raise Exception(f'Data type for {shipment} not recognized')

	def _set_binpack_bin(self,allow_rotations=True,*args,**kwargs):
		'''
		Set Bin Packing Object
		'''
		if not self._binpack_bin_exists:
			width = kwargs.get('width',self.width)
			height = kwargs.get('height',self.height)
			length = kwargs.get('length',self.length)
			max_weight = kwargs.get('max_weight',self.max_weight)
			if allow_rotations:
				self.binpack_bin = Bin(
					name = kwargs.get('name',getattr(self,'name','Unnamed Shipment Object')),
					width = width, 
					height = length,
					depth = height,
					max_weight = max_weight
				)
			else:
				self.binpack_bin = NoRotateBin(
					name = kwargs.get('name',getattr(self,'name','Unnamed Shipment Object')),
					width = width, 
					height = length,
					depth = height,
					max_weight = max_weight
				)
			self._binpack_bin_exists = True
	
	def _delete_binpack_bin(self):
		'''
		Delete Bin Pack Item
		'''
		if hasattr(self,'binpack_item'):
			delattr(self,'binpack_item')
		self._binpack_bin_exists = False
	
	def _update_binpack_bin(self,allow_rotations=True,*args,**kwargs):
		'''
		Delete and Recreate Bid Pack Item
		'''
		self._delete_binpack_bin()
		self._set_binpack_bin(allow_rotations=allow_rotations,*args,**kwargs)
	
	def arrangement_is_valid(self,shipment=None):
		'''
		Validate that the Arrangement of Shipments in Trailer is Physically Possible
		'''
		trailer_boundaries = np.array([[0,self.length],[0,self.width],[0,self.height]])
		if shipment is None:
			shipment_boundaries_list = [shipment.get_boundaries() for shipment in self.shipments]
			for i in range(len(shipment_boundaries_list)):
				s_i = shipment_boundaries_list[i]
				# Validate shipment is inside the trailer
				in_trailer = all(
					[interval_is_subset(s_i[k],trailer_boundaries[k]) for k in range(3)]
				)
				if not in_trailer:
					return False
				for j in range(i):
					s_j = shipment_boundaries_list[j]
					# Validate shipment does not occupy the same space as any other previous shipment
					overlap = all(
						[intervals_overlap(s_i[k],s_j[k]) for k in range(3)]
					)
					if overlap:
						return False
			return True
		else:
			shipment_boundaries_list = [s.get_boundaries() for s in self.shipments if s != shipment]
			s_i = shipment.get_boundaries()
			in_trailer = all([interval_is_subset(s_i[k],trailer_boundaries[k]) for k in range(3)])
			if not in_trailer:
				return False
			for s_j in shipment_boundaries_list:
				overlap = all([intervals_overlap(s_i[k],s_j[k]) for k in range(3)])
				if overlap:
					return False
			return True

	def balance(self):
		'''
		Move Shipments to the Side Walls of the Trailer
		'''
		trailer_center = self.width / 2
		for shipment in self.shipments:
			boundaries = shipment.get_boundaries()
			center = np.mean(boundaries[1,:])
			if center < trailer_center:
				shipment.slide_left(self)
			else:
				shipment.slide_right(self)

	def plot(self):
		'''
		Plot an Overview of the Trailer Load Plan
		'''
		piece_plot_kwargs = {
			'color' : 'red',
			'h_padding' : -2,
			'v_padding' : -2,
		}
		shipment_plot_kwargs = {
			'color' : 'blue',
			'h_padding' : -1,
			'v_padding' : 0,
		}

		ll = self.position
		ur = [self.length,self.width,self.height]
		x_list,y_list = get_rectangle_plot_lists(ll=ll,ur=ur,x_padding=3,y_padding=3)

		pyplot.figure(figsize=(self.length/25,self.width/25))
		pyplot.plot(x_list,y_list,color='black',linewidth=6)
		for shipment in self.shipments:
			s_ll = shipment.position
			for piece in shipment.pieces:
				piece.plot(h_offset=s_ll,v_offset=s_ll,linewidth=2,**piece_plot_kwargs)
				piece.fill(h_offset=s_ll,v_offset=s_ll,alpha=.2,**piece_plot_kwargs)
			shipment.plot(h_offset=ll,v_offset=ll,linewidth=2,**shipment_plot_kwargs)
			shipment.fill(h_offset=ll,v_offset=ll,alpha=.2,**shipment_plot_kwargs)
			boundaries = shipment.get_boundaries()
			x = .5 * boundaries[0,0] + .5 * boundaries[0,1]
			y = .5 * boundaries[1,0] + .5 * boundaries[1,1]
			shipment_str = str(len(shipment.pieces))
			pyplot.text(x,y,shipment_str)
		pyplot.title('Trailer Load Plan',fontsize=20)
		pyplot.xticks(
			[i for i in range(0,int(self.length)+1,12*4)],
			[int(i/12) for i in range(0,int(self.length)+1,12*4)]
		)
		pyplot.xlabel('Feet',fontsize=18)
		pyplot.ylabel('Inches',fontsize=18)
		pyplot.show()

	def get_linear_feet(self):
		'''
		Get Linear Feet for Trailer Load Plan
		'''
		boundaries = self.get_boundaries()
		return DIMENSION_CONVERTER[self.default_dimension_units][options.DimUomFeet](boundaries[0,1] - boundaries[0,0])

	def _set_weight(self):
		'''
		Calculate and Update Trailer Total Weight
		'''
		if len(self.shipments) == 0:
			self.weight = 0
		else:
			self.weight = sum([shipment.weight for shipment in self.shipments])

	def get_load_order(self):
		'''
		Get Load Plan in the Order They Need to be Loaded
		'''
		load_order = {}
		idx = 1
		for shipment in sorted(self.shipments,key=lambda x : x.position[0] * 10000 + x.position[1]):
			for piece in sorted(shipment.pieces,key=lambda x : x.position[2]):
				load_order.update({
					idx : {
						'name' : piece.name,
						'position' : list(shipment.position + piece.position),
						'piece_width' : piece.width,
						'piece_length' : piece.length,
						'piece_height' : piece.height,
						'piece_weight' : piece.weight,
						'piece_packing' : piece.packing,
						'piece_is_rotated' : piece.is_rotated,
						'piece_is_stacked' : piece.position[2] > 0,
					}
				})
				idx += 1
		return load_order

	def get_summary(self):
		'''
		Get Summary of Optimized Trailer Load Plan
		'''
		num_pieces = sum([len(shipment.pieces) for shipment in self.shipments])
		volume = 0
		for shipment in self.shipments:
			for piece in shipment.pieces:
				volume += piece.length * piece.width * piece.height

		actual_cube = 0
		for shipment in self.shipments:
			shipment._set_dims()
			actual_cube += sum([piece.length * piece.width * piece.height for piece in shipment.pieces])

		stacked_cube = sum([shipment.length * shipment.width * shipment.height for shipment in self.shipments])
		trailer_cube = sum([shipment.length * shipment.width * self.height for shipment in self.shipments])

		linear_feet = self.get_linear_feet()

		linear_feet_portion_of_trailer = DIMENSION_CONVERTER[options.DimUomFeet][self.default_dimension_units](linear_feet) / self.length
		actual_cube_portion_of_trailer = actual_cube / self.length / self.width / self.height
		stacked_cube_portion_of_trailer = stacked_cube / self.length / self.width / self.height
		trailer_cube_portion_of_trailer = trailer_cube / self.length / self.width / self.height
		weight_portion_of_trailer = self.weight / self.max_weight

		trailer_stats = {
			'trailer_max_weight' : self.max_weight,
			'trailer_length' : self.length,
			'trailer_width' : self.width,
			'trailer_height' : self.height,
			'arrangement_is_valid' : self.arrangement_is_valid(),
			'trailer_is_overweight' : self.is_overweight(),
			'trailer_capacity_cube' : round(VOLUME_CONVERTER[options.VolumeUomCubicInches][options.VolumeUomCubicFeet](self.length * self.width * self.height),1),

			'num_pieces' : num_pieces,

			'total_weight' : round(self.weight),
			'linear_feet' : round(linear_feet,2),
			'actual_cube' : round(VOLUME_CONVERTER[options.VolumeUomCubicInches][options.VolumeUomCubicFeet](actual_cube),1),
			'stacked_cube' : round(VOLUME_CONVERTER[options.VolumeUomCubicInches][options.VolumeUomCubicFeet](stacked_cube),1),
			'trailer_cube' : round(VOLUME_CONVERTER[options.VolumeUomCubicInches][options.VolumeUomCubicFeet](trailer_cube),1),

			'weight_portion_of_trailer' : round(weight_portion_of_trailer,3),
			'linear_feet_portion_of_trailer' : round(linear_feet_portion_of_trailer,3),
			'actual_cube_portion_of_trailer' : round(actual_cube_portion_of_trailer,3),
			'stacked_cube_portion_of_trailer' : round(stacked_cube_portion_of_trailer,3),
			'trailer_cube_portion_of_trailer' : round(trailer_cube_portion_of_trailer,3),

			'load_order' : self.get_load_order()
		}
		return trailer_stats

	def is_overweight(self):
		'''
		Check if Loaded Trailer is Overweight
		'''
		self._set_weight()
		return self.weight > self.max_weight
		



STANDARD_TRAILER_DIMS = [
    {
        'name': "Standard 53' Dry-Van",
        'code': 'DV_53',
        'inner_width': 98.5,
        'inner_height': 108,
        'inner_length': 52 * 12 + 6,
        'max_weight': 42500,
    },
    {
        'name': "Standard 48' Dry-Van",
        'code': 'DV_48',
        'inner_width': 98.5,
        'inner_height': 100.6,
        'inner_length': 47 * 12 + 6,
        'max_weight': 13500,
    },
    {
        'name': "Standard 45' Dry-Van",
        'code': 'DV_45',
        'inner_width': 93,
        'inner_height': 100.5,
        'inner_length': 44 * 12 + 1.5,
        'max_weight': 11900,
    },
    {
        'name': "Standard 43' Dry-Van",
        'code': 'DV_43',
        'inner_width': 92.5,
        'inner_height': 100.5,
        'inner_length': 41 * 12 + 6,
        'max_weight': 11900,
    },
    {
        'name': "Standard 40' Dry-Van",
        'code': 'DV_40',
        'inner_width': 92.5,
        'inner_height': 94.5,
        'inner_length': 39 * 12 + 6,
        'max_weight': 11300,
    },
    {
        'name': "28' Dry-Van, Pup - Roll Door",
        'code': 'DV_28_RD',
        'inner_width': 100,
        'inner_height': 107,
        'inner_length': 27 * 12 + 3,
        'max_weight': 22000,
    },
    {
        'name': "28' Dry-Van, Pup - Swing Door",
        'code': 'DV_28_SD',
        'inner_width': 99,
        'inner_height': 108,
        'inner_length': 27 * 12 + 3,
        'max_weight': 22000,
    },
    {
        'name': "20' Shipping Container",
        'code': 'CONT_20',
        'inner_width': 7 * 12 + 8,
        'inner_height': 7 * 12 + 9,
        'inner_length': 19 * 12 + 4,
        'max_weight': 22000,
    },
    {
        'name': "40' Shipping Container",
        'code': 'CONT_40',
        'inner_width': 7 * 12 + 8,
        'inner_height': 7 * 12 + 10,
        'inner_length': 39 * 12 + 6,
        'max_weight': 22000,
    },
    {
        'name': "45' Shipping Container",
        'code': 'CONT_45',
        'inner_width': 7 * 12 + 8,
        'inner_height': 8 * 12 + 10,
        'inner_length': 44 * 12 + 6,
        'max_weight': 22000,
    },
    {
        'name': "48' Shipping Container",
        'code': 'CONT_48',
        'inner_width': 8 * 12 + 2,
        'inner_height': 8 * 12 + 10,
        'inner_length': 47 * 12 + 6,
        'max_weight': 22000,
    },
    {
        'name': "53' Shipping Container",
        'code': 'CONT_53',
        'inner_width': 8 * 12 + 2,
        'inner_height': 8 * 12 + 10,
        'inner_length': 52 * 12 + 3,
        'max_weight': 42500,
    },
    {
        'name': "26' Flatbed Trailer",
        'code': 'FB_26',
        'inner_width': 96,
        'inner_height': 102,
        'inner_length': 26 * 12,
        'max_weight': 55000,
    },
    {
        'name': "40' Flatbed Trailer",
        'code': 'FB_40',
        'inner_width': 96,
        'inner_height': 102,
        'inner_length': 40 * 12,
        'max_weight': 55000,
    },
    {
        'name': "45' Flatbed Trailer",
        'code': 'FB_45',
        'inner_width': 102,
        'inner_height': 102,
        'inner_length': 45 * 12,
        'max_weight': 55000,
    },
    {
        'name': "48' Flatbed Trailer",
        'code': 'FB_48',
        'inner_width': 102,
        'inner_height': 102,
        'inner_length': 48 * 12,
        'max_weight': 55000,
    },
    {
        'name': "Small Straight Truck - 22'",
        'code': 'STRGHT_TRK_22',
        'inner_width': 92,
        'inner_height': 8 * 12,
        'inner_length': 21 * 12,
        'max_weight': 14000,
    },
    {
        'name': "Large Straight Truck - 26'",
        'code': 'STRGHT_TRK_26',
        'inner_width': 96,
        'inner_height': 8 * 12,
        'inner_length': 25 * 12,
        'max_weight': 14000,
    },
]




from typing import List, Dict
from ..logistics_objects import (
	Piece, Shipment, Trailer
)
from ..utils import (
	get_current_trailer_configuration, restore_trailer_configuration
)
from ..exceptions import (
	NoPiecesException,
	PiecesTooLongForServiceException,
	OptimizationFailedServiceException,
	InvalidTrailerDimensionsException,
	InvalidPiecesException,
)
from ..validation import (
	validate_piece_length, validate_pieces, validate_trailer_dims
)
from ..optimizer_functions import (
	optimize_pieces_arrangement, optimize_shipment_arrangement,
)
from .. import options
from ..defaults import (
	OVERWEIGHT_SHIPMENT_THRESHOLD,
	DEFAULT_PIECE_ARRANGEMENT_ALGORITHM,
	DEFAULT_SHIPMENT_ARRANGEMENT_ALGORITHM,
)
from copy import deepcopy

def optimize_trailer_load_plan(shipment_list : List[Dict], trailer_dims : Dict, allow_rotations : bool = True, **kwargs):
	# Validation of Pieces and Trailer Dimensions
	try:
		validate_trailer_dims(trailer_dims=trailer_dims)
	except Exception as e:
		raise InvalidTrailerDimensionsException('Invalid trailer dimensions passed') from e
	if len(shipment_list) == 0:
		raise NoPiecesException('No pieces provided for trailer load optimization')
	try:
		validate_pieces(shipment_list=shipment_list)
	except Exception as e:
		raise InvalidPiecesException from e
	try:
		validate_piece_length(
			shipment_list=shipment_list,
			trailer_dims=trailer_dims,
			allow_rotations=allow_rotations,
		)
	except Exception as e:
		raise PiecesTooLongForServiceException('At least one of the pieces provided is too large to fit on the trailer') from e
	
	# Get default and overriden parameters
	dimension_unit_of_measure = options.DimUomInches
	weight_unit_of_measure = options.WeightUomPounds
	overweight_shipment_threshold = OVERWEIGHT_SHIPMENT_THRESHOLD if kwargs.get('overweight_shipment_threshold') is None else kwargs.get('overweight_shipment_threshold')
	piece_arrangement_algorithm = DEFAULT_PIECE_ARRANGEMENT_ALGORITHM if kwargs.get('piece_arrangement_algorithm') is None else kwargs.get('piece_arrangement_algorithm')
	shipment_optimization_ls = DEFAULT_SHIPMENT_ARRANGEMENT_ALGORITHM if kwargs.get('shipment_optimization_ls') is None else kwargs.get('shipment_optimization_ls')
	
	# Create pieces for all provided shipments
	pieces = []
	for i, shipment_dict in enumerate(deepcopy(shipment_list)):
		num_items = shipment_dict.pop('num_pieces', 1)
		for j in range(num_items):
			shipment_dict.update(
				{
					'name': f'Shipment {i+1}:  Piece {j+1}',
					'id' : shipment_dict.get('id'),
					'desc' : shipment_dict.get('desc'),
				}
			)
			pieces += [Piece(**shipment_dict)]
	# Identify overweight shipments
	overweight_pieces = [piece for piece in pieces if piece.weight > overweight_shipment_threshold]
	pieces = [piece for piece in pieces if not piece in overweight_pieces]
	
	# Orient overweight shipments to maximally distribute weight in the length dimension, if allowed
	if allow_rotations:
		for piece in overweight_pieces:
			if piece.length < piece.width:
				piece.rotate()
	
	# Arrange non-overweight pieces into shipments
	if len(pieces) > 0:
		try:
			shipments = optimize_pieces_arrangement(
				pieces=pieces,
				algorithm=piece_arrangement_algorithm,
				trailer_height=trailer_dims.get('inner_height'),
				allow_rotations=allow_rotations,
			)
		except Exception as e:
			raise OptimizationFailedServiceException('Optimization failed during arrangment of pieces into shipments') from e
	else:
		shipments = []
	
	# Create trailer object associated to shipments made up of non-overweight pieces
	trailer = Trailer(
		shipments=shipments,
		width=trailer_dims.get('inner_width'),
		length=trailer_dims.get('inner_length'),
		height=trailer_dims.get('inner_height'),
		dimension_unit_of_measure=dimension_unit_of_measure,
		weight=0,
		weight_unit_of_measure=weight_unit_of_measure,
		max_weight=trailer_dims.get('max_weight'),
	)

	# Optimize shipment arrangement within trailer
	results = []
	if len(pieces) > 0:
		try:
			for opt_kwargs in shipment_optimization_ls:
				optimize_shipment_arrangement(
					trailer=trailer,
					allow_rotations=allow_rotations,
					**opt_kwargs,
				)
				trailer._set_weight()
				results += [
					{
						'linear_feet' : trailer.get_linear_feet(),
						'configuration' : get_current_trailer_configuration(trailer),
						'valid' : trailer.arrangement_is_valid(),
						'request_kwargs' : opt_kwargs,
					}
				]
			assert len([result for result in results if result.get('valid')]) > 0
			result = sorted([result for result in results if result.get('valid')],key=lambda x : x.get('linear_feet'))[0]
			restore_trailer_configuration(result['configuration'])
		except Exception as e:
			raise OptimizationFailedServiceException('Optimization failed during shipment arrangement optimization') from e
	
	# Handle overweight pieces/shipments
	overweight_shipments = [Shipment(pieces=piece) for piece in overweight_pieces]
	boundaries = trailer.get_boundaries()
	try:
		linear_inches = boundaries[0, 1]
		for shipment in overweight_shipments:
			shipment._set_dims()
			shipment.position[0] = linear_inches
			shipment.position[1] = trailer.width / 2 - shipment.width / 2
			trailer.add_shipment(shipment)
			linear_inches += shipment.length
			shipment._set_dims()
			trailer._set_weight()
	except Exception as e:
		raise OptimizationFailedServiceException('Optimization failed during overwieght piece load optimization') from e
	return trailer




OVERWEIGHT_SHIPMENT_THRESHOLD = 2000
DEFAULT_PIECE_ARRANGEMENT_ALGORITHM = 'GREEDY_STACK'
DEFAULT_SHIPMENT_ARRANGEMENT_ALGORITHM = [
    {
        'algorithm' : 'NO_STACK_BIN_PACK',
        'max_iter' : None,
        'timeout' : None,
    },
    {
        'algorithm' : 'SLIDE_BACK',
        'max_iter' : 2,
        'timeout' : 2,
    },
    {
        'algorithm' : 'GREEDY_LOAD',
        'max_iter' : 3,
        'timeout' : None,
    },
]



# -*- coding: utf-8 -*-
from setuptools import setup

install_requires = ['numpy', 'matplotlib', ]

setup(
    name='ytl',
    version='1.0.1',
    author='Jarod Hart & Robert Goss',
    author_email='jhart@yat.ai',
    packages=[
        'ytl', 'ytl.py3dbp', 'ytl.services', 'ytl.logistics_objects',
        'ytl.optimizer_functions',
        'ytl.optimizer_functions.shipment_arrangement',
        'ytl.optimizer_functions.piece_arrangement',
    ],
    license="MIT",
    install_requires=install_requires,
    test_suite="tests",
    classifiers=[
        'Intended Audience :: Developers',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
    ],
    description='YAT Trailer Loading Package',
    long_description=open('README.md').read(),
    zip_safe=True,
)



from decimal import Decimal
from .constants import Axis

def rect_intersect(item1, item2, x, y):
    d1 = item1.get_dimension()
    d2 = item2.get_dimension()

    cx1 = item1.position[x] + d1[x]/2
    cy1 = item1.position[y] + d1[y]/2
    cx2 = item2.position[x] + d2[x]/2
    cy2 = item2.position[y] + d2[y]/2

    ix = max(cx1, cx2) - min(cx1, cx2)
    iy = max(cy1, cy2) - min(cy1, cy2)

    return ix < (d1[x]+d2[x])/2 and iy < (d1[y]+d2[y])/2


def intersect(item1, item2):
    return (
        rect_intersect(item1, item2, Axis.WIDTH, Axis.HEIGHT) and
        rect_intersect(item1, item2, Axis.HEIGHT, Axis.DEPTH) and
        rect_intersect(item1, item2, Axis.WIDTH, Axis.DEPTH)
    )


def get_limit_number_of_decimals(number_of_decimals):
    return Decimal('1.{}'.format('0' * number_of_decimals))


def set_to_decimal(value, number_of_decimals):
    number_of_decimals = get_limit_number_of_decimals(number_of_decimals)

    return Decimal(value).quantize(number_of_decimals)




from typing import List
from ...logistics_objects import Shipment
from ...utils import list_allocate
import numpy as np

greedy_stack_piece_arrangement_details = {
	'code' : 'GREEDY_STACK',
	'name' : 'Greedy Stacking Piece Arrangement',
	'desc' : 'Greedy stacking algorithm to stack smaller pieces on top of larger pieces to make shipments',
}

def try_to_stack(shipment, idx : int, trailer_height : float, allocated_pieces : List[int], available_pieces : List[int], stackable_pieces : List, allow_rotations : bool=True):
	'''
	Try to Stack Piece

	Parameters
	------------
	shipment : logistics_objects.Shipment
		Shipment object attemping to be stacked on
	idx : int
		Index of piece proposed to be stacked on shipment
	trailer_height : float
		Height of trailer
	allocated_pieces : List[int]
		List of indices for pieces already allocated to shipments
	available_pieces : List[int]
		List of indices for pieces unallocated and available to be added to shipments
	stackable_pieces : List[logistics_objects.Shipment]
		List of pieces that are stackable
	allow_rotations : bool
		Allow rotation of pieces when attempting to stack
	
	Returns
	------------
	added_piece_to_shipment : bool
		True if piece was added to shipment, False if not
	'''
	if idx in allocated_pieces:
		return False
	piece = stackable_pieces[idx]
	can_stack = shipment.can_stack(
		piece=piece,
		trailer_height=trailer_height
	)
	if can_stack:
		shipment.stack(piece)
		list_allocate(idx, allocated_pieces, available_pieces)
		return True
	elif allow_rotations:
		piece.rotate()
		can_stack = shipment.can_stack(
			piece=piece,
			trailer_height=trailer_height
		)
		if can_stack:
			shipment.stack(piece)
			list_allocate(idx, allocated_pieces, available_pieces)
			return True
		else:
			piece.rotate()
	return False


def greedy_stack_pieces(pieces : List, trailer_height : int, allow_rotations : bool=True, **kwargs):
	'''
	Allocate Provided Pieces into Shipments Using a Simple Greedy Stacking Algorithm

	Parameters
	------------
	pieces : List[logistics_objects.Piece]
		List of pieces to be allocated into shipments
	trailer_height : int
		Trailer height
	allow_rotations : bool
		Allows pieces to be rotated if True, does not allow rotation otherwise.
	
	Returns
	-----------
	shipments : List[logistics_objects.Shipment]
		List of shipments composed of provided piece objects
	'''
	unstackable_pieces = [p for p in pieces if p.stack_limit <= 1]
	stackable_pieces = [p for p in pieces if p.stack_limit > 1]

	shipments = []
	allocated_pieces = []
	available_pieces = list(range(len(stackable_pieces)))

	sorted_indices = np.argsort(
		[p.length ** .5 + p.width ** .5 for p in stackable_pieces]
	)[::-1]
	count = 0
	max_iter = len(available_pieces) + 1
	while len(available_pieces) > 0 and count < max_iter:
		bottom_idx = [i for i in sorted_indices if i in available_pieces][0]
		list_allocate(bottom_idx, allocated_pieces, available_pieces)
		shipment = Shipment(pieces=stackable_pieces[bottom_idx])
		sorted_indices = np.argsort(
			[p.length ** .5 + p.width ** .5 for p in stackable_pieces]
		)[::-1]
		for idx in sorted_indices:
			_ = try_to_stack(
				shipment=shipment,
				idx=idx,
				trailer_height=trailer_height,
				allocated_pieces=allocated_pieces,
				available_pieces=available_pieces,
				stackable_pieces=stackable_pieces,
				allow_rotations=allow_rotations
			)
		shipment._set_dims()
		shipments += [shipment]
	for piece in unstackable_pieces:
		shipment = Shipment(pieces=piece)
		shipment._set_dims()
		shipments += [shipment]
	return shipments




naive_shipment_arrangement_details = {
	'code' : 'NAIVE',
	'name' : 'Naive Shipment Arrangement',
	'desc' : 'The trivial shipment arrangement - put each shipment in the trailer end to end',
}

def naive_shipment_arrangement(trailer, **kwargs):
	'''
	Arrange Shipment in a Naive Packing, End to End without Optimization
	
	Parameters
	------------
	trailer : logistics_objects.Trailer
		Trailer object to have shipment arranged
	
	Returns
	------------
	None
	'''
	length_pos = 0
	for shipment in trailer.shipments:
		shipment.position[0] = length_pos
		length_pos += shipment.length

